/*
 * Copyright 2009
 *     PYRASIS.COM,.  All rights reserved.
 *
 * Author:
 *     Lee Jae-Hong (pyrasis)
 */


#include <ntddk.h>
#include "condvar.h"


VOID
KeInitializeConditionVariable (
    __out PKCONDITION_VARIABLE ConditionVariable
    )
{
    ConditionVariable->Waiters = 0;
    KeInitializeSpinLock(&ConditionVariable->SpinLock);
    KeInitializeSemaphore(&ConditionVariable->Semaphore, 0, MAXIMUM_WAIT_CONDITION_VARIABLES);
}

NTSTATUS
KeSleepConditionVariableMX (
    __inout PKCONDITION_VARIABLE ConditionVariable,
    __inout PKMUTEX Mutex,
    __in_opt PLARGE_INTEGER Timeout
    )
{
    NTSTATUS status;
    KIRQL irql = KeGetCurrentIrql();

    KeAcquireSpinLock(&ConditionVariable->SpinLock, &irql);
    ConditionVariable->Waiters++;
    KeReleaseSpinLock(&ConditionVariable->SpinLock, irql);

    KeReleaseMutex(Mutex, TRUE);
    
    status = KeWaitForSingleObject(&ConditionVariable->Semaphore, Executive, KernelMode, FALSE, Timeout);

    KeWaitForSingleObject(Mutex, Executive, KernelMode, FALSE, NULL);

    return status;
}

NTSTATUS
KeSleepConditionVariableRS (
    __inout PKCONDITION_VARIABLE ConditionVariable,
    __inout PERESOURCE Resource,
    __in_opt PLARGE_INTEGER Timeout,
    __in ULONG Flags
    )
{
    NTSTATUS status;
    KIRQL irql = KeGetCurrentIrql();

    KeAcquireSpinLock(&ConditionVariable->SpinLock, &irql);
    ConditionVariable->Waiters++;
    KeReleaseSpinLock(&ConditionVariable->SpinLock, irql);

    ExReleaseResourceLite(Resource);

    status = KeWaitForSingleObject(&ConditionVariable->Semaphore, Executive, KernelMode, FALSE, Timeout);

    if (Flags == CONDITION_VARIABLE_LOCKMODE_SHARED)
        ExAcquireResourceSharedLite(Resource, TRUE);
    else
        ExAcquireResourceExclusiveLite(Resource, TRUE);

    return status;
}

VOID
KeWakeConditionVariable (
    __inout PKCONDITION_VARIABLE ConditionVariable
    )
{
    KIRQL irql = KeGetCurrentIrql();

    if (ConditionVariable->Waiters > 0)
    {
        KeAcquireSpinLock(&ConditionVariable->SpinLock, &irql);
        KeReleaseSemaphore(&ConditionVariable->Semaphore, IO_NO_INCREMENT, 1, FALSE);
        ConditionVariable->Waiters--;
        KeReleaseSpinLock(&ConditionVariable->SpinLock, irql);
    }
}

VOID
KeWakeAllConditionVariable (
    __inout PKCONDITION_VARIABLE ConditionVariable
    )
{
    KIRQL irql = KeGetCurrentIrql();

    if (ConditionVariable->Waiters > 0)
    {
        KeAcquireSpinLock(&ConditionVariable->SpinLock, &irql);
        KeReleaseSemaphore(&ConditionVariable->Semaphore, IO_NO_INCREMENT, ConditionVariable->Waiters, FALSE);
        ConditionVariable->Waiters = 0;
        KeReleaseSpinLock(&ConditionVariable->SpinLock, irql);
    }
}
